/*****************************************************************************
* amf_usb_core.c  Amfeltec USB Driver
* 		
* December 2013		
*****************************************************************************/

#ifndef __AMF_USB_CORE__
#define __AMF_USB_CORE__


#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>

#include "amf_usb.h"
#include "amf_usb_utils.h"


//=========================================================================================  SYNC

int amf_usb_syncverify(amf_usb_t *ausb, int len)
{
	unsigned char	data;

	A_BUG(ausb == NULL);

	spin_lock(&ausb->lock);
	if (!ausb->rx_sync){
	
		A_DEBUG_USB("%s: Searching for sync from %d...len %d\n", AMF_USB_NAME, ausb->next_read_ind, len);
		do {
			if ((ausb->readbuf[ausb->next_read_ind] & 0x30) != 0x30){

				ausb->next_read_ind++;
				if(ausb->next_read_ind == MAX_READ_BUF_LEN) {
					ausb->next_read_ind = 0;
				}
				len--;
				continue;
			}
			/* Found first sync condition */
			A_DEBUG_USB("%s: Got first sync condition at %d:%d (%02X)!\n", AMF_USB_NAME, 
				ausb->next_read_ind, ausb->next_rx_ind,
				(unsigned char)ausb->readbuf[ausb->next_read_ind]);
				
			if (len < 4) break;
			
			if ((MAX_READ_BUF_LEN - ausb->next_read_ind) > 4) {
				data = AMF_USB_CTRL_DECODE(&ausb->readbuf[ausb->next_read_ind]);
				if (data == 0x7E){
					/* Found second sync condition */
					A_DEBUG_USB("%s: Got in sync at %d:%d! len %d\n", AMF_USB_NAME, ausb->next_read_ind, ausb->next_rx_ind, len);
					ausb->rx_sync = 1;
					break;
				}
				
				ausb->next_read_ind++;
				if(ausb->next_read_ind == MAX_READ_BUF_LEN) {
					ausb->next_read_ind = 0;
				}
				len--;
			}


		}while(len);

	} else {
		if ((ausb->readbuf[ausb->next_read_ind] & 0x30) != 0x30){
			A_DEBUG_USB("%s: WARNING: Got out of sync 1 at %d:%d (%02X)!\n", AMF_USB_NAME,
						ausb->next_read_ind, ausb->next_rx_ind,
						(unsigned char)ausb->readbuf[ausb->next_read_ind]);
			ausb->stats.rx_sync_err_cnt++;
	
			ausb->rx_sync = 0;
			ausb->next_read_ind++;
			if(ausb->next_read_ind == MAX_READ_BUF_LEN) {
				ausb->next_read_ind = 0;
			}

			spin_unlock(&ausb->lock);
			return 0;
		}


		if ((MAX_READ_BUF_LEN - ausb->next_read_ind) > 4) {
			data = AMF_USB_CTRL_DECODE(&ausb->readbuf[ausb->next_read_ind]);
			if (data != 0x7E){
				A_DEBUG_USB("%s: WARNING: Got out of sync 2 at %d:%d!\n", AMF_USB_NAME, ausb->next_read_ind, ausb->next_rx_ind);
				ausb->stats.rx_sync_err_cnt++;
				ausb->rx_sync = 0;
	
				ausb->next_read_ind++;
				if(ausb->next_read_ind == MAX_READ_BUF_LEN) {
					ausb->next_read_ind = 0;
				}
	
				spin_unlock(&ausb->lock);
				return 0;
			}
		}


		if ((MAX_READ_BUF_LEN - ausb->next_read_ind) > 1) {
	
			if ((ausb->readbuf[ausb->next_read_ind + 1] & 0x20) == 0x20){
				if (!ausb->tx_sync){
					A_DEBUG_USB("%s: USB device is in tx sync! next_read_ind + 1: %d  (%02X)\n", AMF_USB_NAME, 
								ausb->next_read_ind + 1,
								(unsigned char)ausb->readbuf[ausb->next_read_ind + 1]);
				}
				ausb->tx_sync = 1;
			} else {
				if (ausb->tx_sync){
					A_DEBUG_USB("%s: USB device is out of tx sync! next_read_ind + 1: %d  (%02X)\n", AMF_USB_NAME, 
								ausb->next_read_ind + 1,
								(unsigned char)ausb->readbuf[ausb->next_read_ind + 1]);
					ausb->stats.dev_sync_err_cnt++;
				}
				ausb->tx_sync = 0;
			}
		}

	}

	spin_unlock(&ausb->lock);
	return 1;
}



int amf_usb_search_rxsync(amf_usb_t *ausb)
{
	int	retry = 0;
	do{
		if (retry++ > AMF_USBFXO_SYNC_RX_RETRIES) return -EINVAL;
		wait_just_a_bit(AMF_USBFXO_SYNC_DELAY, 0);
	} while(!ausb->rx_sync);
	return 0;
}

int amf_usb_search_txsync(amf_usb_t *ausb)
{
	int		retry = 0;
	do{
		if (retry++ > AMF_USBFXO_SYNC_TX_RETRIES) return -EINVAL;
		wait_just_a_bit(AMF_USBFXO_SYNC_DELAY, 0);
	} while(!ausb->tx_sync);
	return 0;
}

//=========================================================================================  SYNC.   



//=========================================================================================  COMPLETION  ===

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
void amf_read_complete(struct urb *q, struct pt_regs *regs)
#else
void amf_read_complete(struct urb *q)
#endif
{
	struct amf_urb	*aurb = (struct amf_urb*)q->context;
	amf_usb_t *ausb = 		(amf_usb_t*)aurb->pvt;

	int		next_rx_ind = 0;
	int		next_read_ind = 0;
	int		err, len = 0;
	int		i;

	A_BUG(ausb == NULL);

	if (q->status){
		if (!(q->status == -ENOENT ||q->status == -ECONNRESET || q->status == -ESHUTDOWN)){
			A_DEBUG_USB("%s: WARNING: RX transaction failed (%d)!\n",	AMF_USB_NAME, q->status);
		}
		if(q->status == -EPIPE) {
			/* Can't call usb_clear_halt while in_interrupt */
			ausb->stats.tx_endpointstall_cnt++;
			goto read_complete_done;
		}
		return;
	}

	ausb->rxcount++;

	if (!amf_test_bit(AMF_USB_STATUS_READY, &ausb->status)){
		A_DEBUG_USB("%s: WARNING: RX USB core is not ready (%ld:%ld)!\n", AMF_USB_NAME, 
						ausb->rxcount,(unsigned long)jiffies);               
		ausb->stats.core_notready_cnt++;
		return;
	}

	A_DEBUG_RX("%s:%d: [RX:%ld]: RX:%d:%02X %d:%02X %d:%02X (%ld)\n", 
			AMF_USB_NAME, aurb->id, ausb->rxcount, 
			q->actual_length, ((unsigned char*)q->transfer_buffer)[0], 
			ausb->next_read_ind, (unsigned char)ausb->readbuf[ausb->next_read_ind],
			ausb->next_rx_ind, (unsigned char)ausb->readbuf[ausb->next_rx_ind],
			jiffies);

	if (!q->actual_length) goto read_complete_done;

	spin_lock(&ausb->lock);
	next_rx_ind = ausb->next_rx_ind;
	next_read_ind = ausb->next_read_ind;

	for (i = 0; i < q->actual_length; i++) {
		ausb->readbuf[next_rx_ind] = ausb->readbuf_block[i];

		next_rx_ind++;
		if (next_rx_ind == MAX_READ_BUF_LEN){
			next_rx_ind = 0;
		}

	}

	/* analyze data */ 
	if (next_rx_ind >= next_read_ind){
		len = next_rx_ind - next_read_ind; 
	}else{
		len = MAX_READ_BUF_LEN - next_read_ind + next_rx_ind;
	}

	ausb->next_rx_ind = next_rx_ind;
	ausb->next_read_ind = next_read_ind;
	spin_unlock(&ausb->lock);	

	if (len > 4) {
		amf_usb_syncverify(ausb, len);
	}

	if ((ausb->rx_sync) && (len >= MAX_USB_RX_LEN)){
		tasklet_schedule(&ausb->bh_task);
	}


read_complete_done:

	q->transfer_buffer = &ausb->readbuf_block[0];
	if ((err = amf_usb_start_transfer(aurb))){
		if (err != -ENOENT){
			A_DEBUG("%s: Failed to program receiver ( )!\n", AMF_USB_NAME); 
		}
	}

	return;

}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
void amf_write_complete(struct urb *q, struct pt_regs *regs)
#else
void amf_write_complete(struct urb *q)
#endif
{
	struct amf_urb	*aurb = (struct amf_urb*)q->context;
	amf_usb_t		*ausb =	(amf_usb_t*)aurb->pvt;
	int 	actual_length =	MAX_USB_TX_LEN;

	A_BUG(ausb == NULL);

	ausb->txcount++;
	if (q->status){
		if (!(q->status == -ENOENT ||q->status == -ECONNRESET || q->status == -ESHUTDOWN || q->status == -EPIPE)){
			A_DEBUG("%s: WARNING: TX transaction failed (%d)!\n", AMF_USB_NAME, q->status);
		}
		if(q->status == -EPIPE) {
			/* Can't call usb_clear_halt while in_interrupt */
			ausb->stats.rx_endpointstall_cnt++;
		}
		return;
	}

	if (!amf_test_bit(AMF_USB_STATUS_READY, &ausb->status)){
		A_DEBUG_USB("%s: WARNING: TX USB core is not ready (%ld:%ld)!\n", AMF_USB_NAME, ausb->txcount,	jiffies);
		ausb->stats.core_notready_cnt++;
		return;
	}

	if (q->actual_length < MAX_USB_TX_LEN){
		A_DEBUG_USB("%s: WARNING: Tx length < MAX(%d:%d)\n", AMF_USB_NAME, q->actual_length, MAX_USB_TX_LEN);
		actual_length = q->actual_length; 
	}

	/* Mark this urb as ready */
	amf_set_bit(AMF_URB_STATUS_READY, &aurb->ready);
	amf_set_bit(AMF_USB_STATUS_TX_READY, &ausb->status);

	return;

}

//========================================================================== COMPLETION. =======================



//======================================================================================   ENDPOINTS & URBS

void amf_usb_set_transfer_urb(struct usb_interface *interface) {

	struct usb_host_interface *iface_desc; 
	struct usb_endpoint_descriptor *endpoint;
	struct amf_usb *ausb;   
	int i;
	int x;

	ausb = usb_get_intfdata(interface);

	iface_desc = interface->cur_altsetting;

	for (i = 0; i < iface_desc->desc.bNumEndpoints; i++) {
		endpoint = &iface_desc->endpoint[i].desc;	

		if (usb_endpoint_is_bulk_in(endpoint)) {
			for (x = 0; x < MAX_READ_URB_COUNT; x++) {
				
				usb_init_urb(&ausb->urbread[x].urb);
				usb_fill_bulk_urb (
						&ausb->urbread[x].urb,
						ausb->udev,
						usb_rcvbulkpipe (ausb->udev, endpoint->bEndpointAddress),
						&ausb->readbuf_block[64], 
						endpoint->wMaxPacketSize,                    			
						amf_read_complete, 										
						&ausb->urbread[x]);	//p									

				A_DEBUG("%s: Bulk IN EP Address:%X Len:%d\n",
						AMF_USB_NAME, 
						endpoint->bEndpointAddress,
						endpoint->wMaxPacketSize);
			}
		}				

		if (usb_endpoint_is_bulk_out(endpoint)) {
			for (x = 0; x < MAX_WRITE_URB_COUNT; x++) {

				usb_init_urb(&ausb->urbwrite[x].urb);
				usb_fill_bulk_urb (
						&ausb->urbwrite[x].urb,
						ausb->udev,
						usb_sndbulkpipe (ausb->udev, endpoint->bEndpointAddress),
						&ausb->writebuf[MAX_USB_TX_LEN*x], 
						MAX_USB_TX_LEN,                          				 
						amf_write_complete,										
						&ausb->urbwrite[x]);	//p								
																				
				ausb->urbwrite[x].next_off = MAX_USB_TX_LEN*x;
				A_DEBUG("%s: Bulk OUT EP Address:%X Len:%d\n", 	 
						AMF_USB_NAME, 
						endpoint->bEndpointAddress,
						MAX_USB_TX_LEN);
			}
		}
	}
	return;
}



//======================================================================================   URB STATUS READY
void amf_usb_set_urb_status_ready(amf_usb_t *ausb) {
	int x;

	ausb->urbcount_read = MAX_READ_URB_COUNT;	//4
	ausb->urb_read_ind = 0;
	for (x = 0; x < ausb->urbcount_read; x++){
		ausb->urbread[x].id = x;
		ausb->urbread[x].pvt = ausb;
		amf_set_bit(AMF_URB_STATUS_READY, &ausb->urbread[x].ready);
	}

	ausb->urbcount_write = MAX_WRITE_URB_COUNT;	//4
	ausb->urb_write_ind = 0;
	for (x = 0; x < ausb->urbcount_write; x++){
		ausb->urbwrite[x].id = x;
		ausb->urbwrite[x].pvt = ausb;
		amf_set_bit(AMF_URB_STATUS_READY, &ausb->urbwrite[x].ready);
	}
	return;
}


//======================================================================================   Buffers ALLOC
int amf_usb_rxtx_buffers_alloc(amf_usb_t *ausb) {

	int size;
	int x;
	struct skb_buff	*skb;
	int err = -ENOMEM;
		

	size = MAX_WRITE_BUF_LEN + MAX_USB_TX_LEN + 1;
	ausb->writebuf = kmalloc(size, GFP_ATOMIC);         
	if (ausb->writebuf == NULL){
		A_DEBUG("%s: Failed to allocate TX buffer!\n", AMF_USB_NAME);
		goto error;
	}
	memset(ausb->writebuf,0x00,size);
	amf_set_bit(AMF_USB_LSTATUS_WRITEBUF, &ausb->lstatus);

	//size = MAX_READ_BUF_LEN + MAX_USB_RX_LEN + 1;
	size = MAX_READ_BUF_LEN;
	ausb->readbuf = kmalloc(size, GFP_ATOMIC);          
	if (ausb->readbuf == NULL){
		A_DEBUG("%s: Failed to allocate RX buffer!\n", AMF_USB_NAME);
		goto error;
	}
	memset(ausb->readbuf,0x00,size);
	amf_set_bit(AMF_USB_LSTATUS_READBUF, &ausb->lstatus);

	AMF_QUE_INIT(&ausb->rx_cmd_free_list);
	AMF_QUE_INIT(&ausb->rx_cmd_list);
	for(x = 0; x < AMF_USB_MAX_RX_CMD_QLEN; x++){
		skb = amf_skb_alloc(10);
		if (!skb){
			A_DEBUG("%s: Failed to allocate RX cmd buffer!\n", AMF_USB_NAME);
			goto error;
		}
		amf_skb_queue_tail(&ausb->rx_cmd_free_list, skb);
	}
	amf_set_bit(AMF_USB_LSTATUS_RXCMD, &ausb->lstatus);

	AMF_QUE_INIT(&ausb->tx_cmd_free_list);
	AMF_QUE_INIT(&ausb->tx_cmd_list);
	for(x = 0; x < AMF_USB_MAX_TX_CMD_QLEN; x++){
		skb = amf_skb_alloc(10);
		if (!skb){
			A_DEBUG("%s: Failed to allocate TX cmd buffer!\n", AMF_USB_NAME);
			goto error;
		}
		amf_skb_queue_tail(&ausb->tx_cmd_free_list, skb);
	}
	amf_set_bit(AMF_USB_LSTATUS_TXCMD, &ausb->lstatus);


	return 0;
error:
	return err;
}
//======================================================================================   Buffers ALLOC.


//=====================================  ENCODE / DECODE ===========================================

static int amf_usb_rxcmd_decode(amf_usb_t *ausb, u8 *rx_cmd, int rx_cmd_len)
{
	struct sk_buff	*skb;
	u_int8_t		*cmd;
	u_int8_t		reg_no;
	int				mod_no;

	A_BUG(ausb == NULL);

	switch(AMF_USB_CMD_TYPE_DECODE(rx_cmd[0])){
	case AMF_USB_CMD_TYPE_WRITE_CPU:
		A_DEBUG_RX("%s: Rx Write CPU register (%02X:%02X)!\n",  
					AMF_USB_NAME,
					(unsigned char)rx_cmd[1],
					(unsigned char)rx_cmd[2]);
		break;			
	case AMF_USB_CMD_TYPE_READ_CPU:
		A_DEBUG_RX("%s: Rx Read CPU register (%02X:%02X)!\n",   
					AMF_USB_NAME,
					(unsigned char)rx_cmd[1],
					(unsigned char)rx_cmd[2]);
		break;
	case AMF_USB_CMD_TYPE_WRITE_FXO:
		A_DEBUG_RX("%s: Rx Write FXO register (%02X:%02X)!\n",  
					AMF_USB_NAME,
					(unsigned char)rx_cmd[1],
					(unsigned char)rx_cmd[2]);
		break;
	case AMF_USB_CMD_TYPE_READ_FXO:
		A_DEBUG_RX("%s: Rx Read FXO register (%d:%02X)!\n",     
					AMF_USB_NAME,
					(unsigned char)rx_cmd[1],
					(unsigned char)rx_cmd[2]);
		break;
	case AMF_USB_CMD_TYPE_EVENT:
		mod_no = rx_cmd[0] & 0x01;
		reg_no = rx_cmd[1];
		A_DEBUG_RX("%s: Module %d: Rx Event Info (Reg:%d = %02X:%ld)!\n",   
					AMF_USB_NAME,
					mod_no+1, 
					reg_no, rx_cmd[2],
					jiffies);
		ausb->regs[mod_no][reg_no] = rx_cmd[2];
		return 0;
	default:
		A_DEBUG_RX("%s: Unknown command %X : %02X!\n",  
				AMF_USB_NAME,
				AMF_USB_CMD_TYPE_DECODE(rx_cmd[0]), rx_cmd[0]);
		ausb->stats.rx_cmd_unknown++;
		return -EINVAL;
	}
	skb = amf_skb_dequeue(&ausb->rx_cmd_free_list);
	cmd = amf_skb_put(skb, rx_cmd_len);
	memcpy(cmd, &rx_cmd[0], rx_cmd_len);
	amf_skb_queue_tail(&ausb->rx_cmd_list, skb);
	return 0;
}




static int amf_usb_rxdata_decode (amf_usb_t *ausb, unsigned char *rxdata)
{
	int		x = 0, mod_no=0, ind=0, high=0, shift=0;
	unsigned char	data, cmd, start_bit, start_fr_bit;
	u_int8_t	data_tmp;
	u_int8_t	*rx_data[2];
	u_int8_t	rx_cmd[10];
	int		rx_cmd_len= 0;

	A_BUG(ausb == NULL);

	rx_data[0] = ausb->readchunk[0];
	rx_data[1] = ausb->readchunk[1];

	memset(&rx_cmd[0],0, 10);
	x = 0;

	while(x < MAX_USB_RX_LEN){

		A_DEBUG_RX("  RX: %02X\n", (unsigned char)rxdata[x]);
		data		= rxdata[x] & 0x0F;
		start_bit	= rxdata[x] & 0x10;
		start_fr_bit	= rxdata[x] & 0x20;

		ind	= x / 4;
		mod_no	= x % 2;
		high	= (x % 4) / 2;
		/* decode voice traffic */
		if (mod_no == 0 && high == 0 && (ind % AMF_USB_CHUNKSIZE) == 0){
			if (!start_fr_bit){
				A_DEBUG(
				"%s: WARNING:%ld: Start Frame bit missing (%d:%d:%02X:%ld)\n",
						AMF_USB_NAME,
						ausb->bhcount,
						x,ind,
						(unsigned char)rxdata[x],
						jiffies);
				ausb->stats.rx_start_fr_err_cnt++;
				break;
			}
		}
		if (x % 4 == 0){
			if (!start_bit){
				A_DEBUG(
				"%s: WARNING:%ld: Start bit missing (%d:%02X:%ld)\n",
						AMF_USB_NAME,
						ausb->bhcount,
						x,
						(unsigned char)rxdata[x],
						jiffies);
				ausb->stats.rx_start_err_cnt++;
				break;
			}
		}
		if (high){
			rx_data[mod_no][ind] |= (data << 4);
			data_tmp = rx_data[mod_no][ind];
		}else{
			rx_data[mod_no][ind] = data;
		}


		/* decode control command (2 bits) */
		cmd = rxdata[x] & 0xC0;

		ind = x / 4;
		shift = (x % 4) * 2;
		cmd = cmd >> shift;
		if (ind && shift == 0){
			if (rx_cmd[rx_cmd_len] == ausb->ctrl_idle_pattern){
				A_DEBUG_RX("RX_CMD: x:%d ind:%d shift:%d len:%d ctrl_idle_pattern\n", x, ind, shift, rx_cmd_len);
				rx_cmd[rx_cmd_len] = 0x00;
				if (rx_cmd_len){
					if (!amf_skb_queue_len(&ausb->rx_cmd_free_list)){
						A_DEBUG("%s: WARNING:%ld: Received too many commands!\n", AMF_USB_NAME, jiffies);
						ausb->stats.rx_cmd_drop_cnt++;
						rx_cmd_len = 0;
					}
					if (rx_cmd_len >= 3){
						amf_usb_rxcmd_decode(ausb, rx_cmd, rx_cmd_len);
					}else{
						ausb->stats.rx_cmd_reset_cnt++;
						if (rx_cmd_len == 1){
							A_DEBUG_RX("%s: Reset RX Cmd (%d:%02X:%02X)!\n", AMF_USB_NAME, rx_cmd_len,
										AMF_USB_CMD_TYPE_DECODE(rx_cmd[0]), rx_cmd[0]);
						}else{
							A_DEBUG_RX("%s: Reset RX Cmd (%d:%02X:%02X:%02X)!\n", AMF_USB_NAME, rx_cmd_len,	
										AMF_USB_CMD_TYPE_DECODE(rx_cmd[0]), rx_cmd[0],rx_cmd[1]);
						}
					}
					memset(&rx_cmd[0],0,rx_cmd_len);
					rx_cmd_len = 0;
				}
			}else{
				A_DEBUG_RX("RX_CMD: x:%d ind:%d shift:%d cmd:%02X\n", x, ind, shift, (unsigned char)rx_cmd[rx_cmd_len]);
				rx_cmd_len++;
			}
		}
		rx_cmd[rx_cmd_len] |= cmd;
		A_DEBUG_RX("RX_CMD: x:%d ind:%d shift:%d cmd:%02X:%02X\n",
					x, ind, shift, (unsigned char)cmd,
					(unsigned char)rx_cmd[rx_cmd_len]);
		x++;
	}


	return 0;
}


static int amf_usb_txdata_encode (amf_usb_t *ausb, unsigned char *writebuf)
{
	int		off, x = 0, ind, mod_no, tx_idle = 0;
	unsigned char	data, start_bit;
	unsigned char	*txdata[2];

	A_BUG(ausb == NULL);

	txdata[0] = ausb->writechunk[0];
	txdata[1] = ausb->writechunk[1];
	memcpy(writebuf, &ausb->idlebuf[0], MAX_USB_TX_LEN); 
	for (x = 0; x < AMF_USB_RXTX_CHUNKSIZE; x++) {
		for (mod_no = 0; mod_no < 2; mod_no++) {
			start_bit = 0x00;
			if (tx_idle){
				data = AMF_USB_IDLE_PATTERN;
			}else{
				data = *(txdata[mod_no] + x);
			}
	
			ind = x * 4 + mod_no;
			AMF_USB_CTRL_ENCODE(writebuf, ind);
			/* write loq nibble */
			AMF_USB_DATA_ENCODE(&writebuf[ind], data);
		}
	}	

	if (amf_skb_queue_len(&ausb->tx_cmd_list)){
		u8				*cmd;
		int				cmd_len;
		struct sk_buff	*skb;
	
		skb = amf_skb_dequeue(&ausb->tx_cmd_list);
		cmd = amf_skb_data(skb);
		cmd_len = amf_skb_len(skb);
		off = 0;

		off++;
		for (x=0; x < cmd_len; x++,off++) {
			AMF_USB_CMD_ENCODE(&writebuf[off*4], cmd[x]);
		}
		amf_skb_init(skb, 0);
		amf_skb_queue_tail(&ausb->tx_cmd_free_list, skb);
	}
	return 0;
}


//===================================== ENCODE / DECODE. ===========================================




//==========================================================================    RX_TX_EVENTS   ==========

int amf_usb_rxdata_enable(amf_usb_t *ausb, int enable)
{
	A_BUG(ausb == NULL);

	if (enable){
		amf_set_bit(AMF_USB_STATUS_RX_DATA_READY, &ausb->status);
	}else{
		amf_clear_bit(AMF_USB_STATUS_RX_DATA_READY, &ausb->status);
	}
	return 0;
}


int amf_usb_rxevent_enable(amf_usb_t *ausb, int mod_no, int enable)
{
	int		event_bit;
	u8		mask;

	A_BUG(ausb == NULL);

	if (mod_no) mask = AMF_USB_CPU_BIT_CTRL_TS1_EVENT_EN;
	else mask = AMF_USB_CPU_BIT_CTRL_TS0_EVENT_EN;
	if (enable) ausb->reg_cpu_ctrl |= mask;
	else ausb->reg_cpu_ctrl &= ~mask; 
	
	__amf_usb_cpu_write(ausb, AMF_USB_CPU_REG_CTRL, ausb->reg_cpu_ctrl);

	event_bit = (mod_no) ? AMF_USB_STATUS_RX_EVENT2_READY : AMF_USB_STATUS_RX_EVENT1_READY;
	amf_set_bit(event_bit, &ausb->status);

	return 0;
}

int amf_usb_rxevent(amf_usb_t *ausb, int mod_no, int force)
{
	int		event_bit;

	A_BUG(ausb == NULL);

	event_bit = (mod_no) ? AMF_USB_STATUS_RX_EVENT2_READY : AMF_USB_STATUS_RX_EVENT1_READY;
	if(!test_bit(event_bit, (unsigned long *)&ausb->status)){
		return -EINVAL;
	}
	if (force){

		ausb->regs[mod_no][5]  = __amf_usb_fxo_read(ausb, mod_no, 5);
		ausb->regs[mod_no][29] = __amf_usb_fxo_read(ausb, mod_no, 29);
		ausb->regs[mod_no][34] = __amf_usb_fxo_read(ausb, mod_no, 34);
		ausb->regs[mod_no][4]  = __amf_usb_fxo_read(ausb, mod_no, 4);

	}

	return 0;
}

static int	amf_rx_tx_span(amf_usb_t *ausb) {

	int x;
	int mod_no;

	x = ausb->isrcount % ausb->mod_cnt;

	if (ausb->isrcount % 2 == 0){
		amf_check_hook(ausb, 0);
	}

	if (!(ausb->isrcount & 0xf0)){
		amf_voicedaa_recheck_sanity(ausb, x);
	}

	if (amf_test_bit(AMF_USB_LSTATUS_DAHDI, &ausb->lstatus)){

		for (mod_no = 0; mod_no < ausb->mod_cnt; mod_no++) {
			dahdi_ec_chunk(&ausb->chans[mod_no], ausb->chans[mod_no].readchunk, ausb->lasttx[mod_no]);
		}

		dahdi_receive(&ausb->span);
		dahdi_transmit(&ausb->span);

	}	
	
	for (mod_no = 0; mod_no < ausb->mod_cnt; mod_no++) {
		memcpy(ausb->lasttx[mod_no], ausb->chans[mod_no].writechunk, DAHDI_CHUNKSIZE);
	}

	return 0;
}


//==========================================================================    RX_TX_EVENTS.   ==========



//==========================================================================    ISR   ==========

void amf_usb_isr(void *arg)
{
	amf_usb_t		*ausb = (amf_usb_t*)arg;
	int i;

	ausb->isrcount++;

	for (i = 0; i < ausb->mod_cnt; i++) {
		
		amf_usb_rxevent(ausb, i, 0);

	}

	amf_rx_tx_span(ausb);


	return;
}



//========================================================================== TASKLET ==========

int amf_usb_start_transfer(struct amf_urb *aurb)
{
	int	ret;

	ret = usb_submit_urb(&aurb->urb, GFP_ATOMIC); 
	if (ret){
		A_DEBUG_USB("%s: usb_submit %p returns %d (%s)\n", AMF_USB_NAME, aurb, ret,
					amf_test_bit(AMF_URB_STATUS_READY, &aurb->ready)?"Ready":"Busy");
		return ret;
	}
	return 0;
}


static int amf_usb_txdata_prepare(amf_usb_t *ausb, char *writebuf, int len)
{
	int		next_write_ind, urb_write_ind; 
	int		err;

	A_BUG(ausb == NULL);
	A_BUG(len > MAX_USB_TX_LEN);

	
	next_write_ind = ausb->next_write_ind;

	memcpy(&ausb->writebuf[next_write_ind], writebuf, len); 

	next_write_ind = (next_write_ind + len) % MAX_WRITE_BUF_LEN;  
	if (next_write_ind == ausb->next_tx_ind){
		A_DEBUG_USB("WARNING: %ld: TX BH is too fast (%d:%ld)\n",
					ausb->bhcount,
					next_write_ind, 
					(unsigned long)jiffies);
		ausb->stats.tx_overrun_cnt++;
	}else{
		ausb->next_write_ind = next_write_ind;
	}

	urb_write_ind = (ausb->urb_write_ind + 1) % MAX_WRITE_URB_COUNT;

	if (!amf_test_bit(AMF_URB_STATUS_READY, &ausb->urbwrite[urb_write_ind].ready)){
		/* Use previous urb for now and increment errors */
		A_DEBUG_TX("%s: [BH:%ld:%ld:%ld]: TX is not ready (%ld)!\n",
				AMF_USB_NAME, 
				ausb->bhcount,
				ausb->rxcount,
				ausb->txcount,
				(unsigned long)jiffies);
		ausb->stats.tx_overrun_cnt++;
		return -EINVAL;
	}

	/* Update tx urb */
	ausb->urbwrite[urb_write_ind].urb.transfer_buffer = &ausb->writebuf[ausb->next_tx_ind];
	ausb->urbwrite[urb_write_ind].urb.transfer_buffer_length = len;
	ausb->next_tx_ind += len;
	ausb->next_tx_ind = ausb->next_tx_ind % MAX_WRITE_BUF_LEN;
	amf_clear_bit(AMF_USB_STATUS_TX_READY, &ausb->status);


	if ((err = amf_usb_start_transfer(&ausb->urbwrite[urb_write_ind]))){
		if (err != -ENOENT){
			A_DEBUG("%s: Failed to program transmitter (%ld)!\n", AMF_USB_NAME, jiffies);
		}
		ausb->stats.tx_notready_cnt++;
		return -EINVAL;
	}

	ausb->urb_write_ind = urb_write_ind;

	return 0;
}


static int amf_usb_rxdata_get(amf_usb_t *ausb, unsigned char *rxdata, int max_len)
{
	int		rx_len;
	int 	i;
	int		next_read_ind = ausb->next_read_ind;
	int		next_rx_ind = ausb->next_rx_ind;
	
	
	A_BUG(ausb == NULL);

	if (next_rx_ind > next_read_ind){
		rx_len = next_rx_ind - next_read_ind; 
	}else if (next_rx_ind < next_read_ind){					
		rx_len = MAX_READ_BUF_LEN - next_read_ind + next_rx_ind;
	}else{
		return 0;
	}

	spin_lock(&ausb->lock);

	if (rxdata == NULL){
		/* ignore previously received data */
		next_read_ind = (next_read_ind + rx_len) % MAX_READ_BUF_LEN;
		spin_unlock(&ausb->lock);
		return rx_len;
	}

	if (rx_len > max_len) rx_len = max_len;                         
	if (rx_len > MAX_USB_RX_LEN) rx_len = MAX_USB_RX_LEN;

	for (i = 0; i < rx_len; i++) {

		rxdata[i] = ausb->readbuf[next_read_ind];
		next_read_ind++;
		if (next_read_ind == MAX_READ_BUF_LEN) {
			 next_read_ind = 0;
		}

	}

	ausb->next_read_ind = next_read_ind;
	
	spin_unlock(&ausb->lock);
	return rx_len;

}




static int amf_usb_bh_rx (amf_usb_t *ausb)
{
	int		rx_len;
	u_int8_t	readbuf[MAX_USB_RX_LEN];

	A_BUG(ausb == NULL);

	rx_len = amf_usb_rxdata_get(ausb, &readbuf[0], MAX_USB_RX_LEN);
	if (!rx_len){
		A_DEBUG_USB("%s: WARNING:%ld: No data available!\n", AMF_USB_NAME, jiffies);
		ausb->stats.rx_underrun_cnt++;
		return -EINVAL;
	}

	if (rx_len < MAX_USB_RX_LEN){
		A_DEBUG_USB("%s: WARNING: Not enough data (%d:%d:%d:%d)!\n", AMF_USB_NAME, rx_len,
				ausb->next_read_ind,
				ausb->next_rx_ind,
				MAX_READ_BUF_LEN);
		ausb->stats.rx_underrun_cnt++;
		return 0;
	}

	amf_usb_rxdata_decode(ausb, readbuf);

	if (amf_test_bit(AMF_USB_STATUS_RX_DATA_READY, &ausb->status) && ausb->isr_func){
		ausb->isr_func(ausb->isr_arg);
	}

	return 0;
}


static int amf_usb_bh_tx (amf_usb_t *ausb)
{	
	unsigned char	writebuf[MAX_USB_TX_LEN];

	A_BUG(ausb == NULL);
	
	amf_usb_txdata_encode(ausb, writebuf);

 	if (amf_usb_txdata_prepare(ausb, writebuf, MAX_USB_TX_LEN)){
		return -EINVAL;
	}

	return 0;
}


void amf_usb_bh (unsigned long data)
{
	amf_usb_t		*ausb = (amf_usb_t*)data;
	
	A_BUG(ausb == NULL);

	if (!amf_test_bit(AMF_USB_STATUS_READY, &ausb->status)){
		goto usb_bh_exit;
	}
	
	if (amf_test_and_set_bit(AMF_USB_STATUS_BH, &ausb->status)){
		A_DEBUG("%s: [BH:%ld]: Re-entry in USB BH!\n", AMF_USB_NAME, jiffies);
		goto usb_bh_done;
	}
	ausb->bhcount++;

	if (ausb->failover_enabled) {
		if ((ausb->bhcount % 5000) == 0) {
			A_DEBUG_TX("%s%d: WATCHDOG: %ld\n", AMF_USB_NAME, ausb->order, jiffies);
			__amf_usb_cpu_write(ausb, 0x0A, 0x02);	
		}	
	}

	/* receive path */
	amf_usb_bh_rx(ausb);
	/* transmit path */
	amf_usb_bh_tx(ausb);

	amf_clear_bit(AMF_USB_STATUS_BH, &ausb->status);

usb_bh_done:

usb_bh_exit:
	return;
}

//========================================================================== TASKLET. ==========



//========================================================================== RAW TX/RX =========

int amf_usb_txdata_raw(amf_usb_t *ausb, unsigned char *data, int max_len)
{
	int		err = 0;

	if (max_len > MAX_USB_TX_LEN){
		max_len = MAX_USB_TX_LEN;				// 32
	}

	err = amf_usb_txdata_prepare(ausb, data, max_len);

	return (!err) ? max_len : 0;
}

int amf_usb_txdata_raw_ready(amf_usb_t *ausb)
{
	if (!amf_test_bit(AMF_USB_STATUS_TX_READY, &ausb->status)){
		return -EBUSY;
	}
	return 0;
}

int amf_usb_rxdata_raw(amf_usb_t *ausb, unsigned char *data, int max_len)
{
	int		rxlen = 0;

	if (max_len > MAX_USB_RX_LEN){
		max_len = MAX_USB_RX_LEN;				// 32
	}

	rxlen = amf_usb_rxdata_get(ausb, data, max_len);
	return rxlen;
}


int amf_usb_rxurb_reset(amf_usb_t *ausb)
{
	int		x=0;

	ausb->next_rx_ind	= 0;
	ausb->next_read_ind	= 0;
	for (x = 0; x < ausb->urbcount_read; x++){
		//ausb->urbread[x].urb.transfer_buffer = &ausb->readbuf[ausb->next_rx_ind];
		ausb->urbread[x].urb.transfer_buffer = &ausb->readbuf_block[0];
		if (amf_usb_start_transfer(&ausb->urbread[x])) {
			A_DEBUG("%s: Failed to start RX:%d transfer!\n", AMF_USB_NAME, x);
			return -EINVAL;
		}
	}
	return 0;
}

int amf_usb_txurb_reset(amf_usb_t *ausb)
{

	ausb->next_tx_ind		= 0;
	ausb->next_write_ind	= 0;
	return 0;
}




#endif  //__AMF_USB_CORE__


